#include <sys/time.h>
#include <sys/types.h>
#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <pgm.h>

Display *dpy;
int  screen;
Colormap colormap;

void usage()
{
  fputs("pgmmovie [-color] [-linear] [-framerate #] [-repeat #] pgmfiles\n",stderr);
  exit(1);
}

int timetoint(a)
struct timeval *a;
{
  return ((1000 * a->tv_sec) + (a->tv_usec/1000));
}

void remappixels(p,c,v)
gray *p;
int c;
int v;
{
  while (--c >= 0) {
    *p++ += v;
  }
}

int main(argc,argv)
int argc;
char *argv[];
{
  int i,j;
  Visual *visual = NULL;
  XImage *image =  NULL;
  XSetWindowAttributes attributes;
  XSizeHints sizehints;
  XEvent event;
  Window image_win;
  XGCValues gc_val;
  GC gc;
  char *dispname = NULL;
  Pixmap *images,pxm;
  int startframe;
  int last,left;
  struct timeval now,timeout;
  struct timezone dummy;
  int width,height;
  gray maxval;
  gray *pixels;
  XStandardColormap stdcmap;
  int format;
  int framerate = 15;
  int linear = 0;
  int repeats = 10;
  int colored = 0;		/* Really colored stuff */
	
  pgm_init( &argc, argv);

  for (i = 1; i < argc; i++) {
    if (argv[i][0] == '-') {
      if (strncmp(argv[i], "-repeat", 7) == 0) {
	if (++i < argc) {
          repeats = atoi(argv[i]);
	} else {
	  usage();
	}
	continue;
      }
      if (strncmp(argv[i], "-framerate", 10) == 0) {
	if (++i < argc) {
          framerate = atoi(argv[i]);
	} else {
	  usage();
	}
        continue;
      }
      if (strncmp(argv[i], "-linear", 7) == 0) {
        linear = 1;
        continue;
      }
      if (strncmp(argv[i], "-color", 8) == 0) {
        colored = 1;
        continue;
      }
      usage();
    } else {
      break;
    }
  }

  if (i == argc) {
    fprintf(stderr,"No files to animate ?\n");
    exit(0);
  }

  if (!(dpy = XOpenDisplay(dispname))) {
    fprintf(stderr,"Can't open display %s\n",getenv("DISPLAY"));
    exit(1);
  }
  screen =  DefaultScreen(dpy);
  visual =  DefaultVisual(dpy, screen);

  stdcmap.colormap = 0;		/* So that we catch an error here */
  if (colored) {
    XGetStandardColormap(dpy,RootWindow(dpy, screen),&stdcmap,XA_RGB_DEFAULT_MAP);
    colormap = stdcmap.colormap;
    if (colormap && (stdcmap.red_mult != 25 || stdcmap.green_mult != 5)) {
        fprintf(stderr,"RGB_DEFAULT_MAP isn't a 5x5x5 cube\n");
        fprintf(stderr,"use xstdcmap -default to install one\n");
	exit(1);
    }
    if (!colormap) {
        fprintf(stderr,"use xstdcmap -default to install standard colormap\n");
	exit(1);
    }
  } else {
    XGetStandardColormap(dpy,RootWindow(dpy, screen),&stdcmap,XA_RGB_GRAY_MAP);
    colormap = stdcmap.colormap;
    if (!colormap) {
        fprintf(stderr,"use xstdcmap -gray to install standard colormap\n");
	exit(1);
    }
  }
  startframe = i;
	
  images = (Pixmap*) malloc(sizeof(Pixmap)*argc);
  
  gc_val.graphics_exposures = 0;
  gc = XCreateGC(dpy, RootWindow(dpy, screen), GCGraphicsExposures, &gc_val);

  while (i < argc) {
    FILE *fp;
    gray *data;

    fp = fopen(argv[i],"r");
    pgm_readpgminit(fp, &width, &height, &maxval,&format);
    data = pixels  = (gray*) malloc(width*height);
    for (j = 0 ; j < height; j++) {
      pgm_readpgmrow(fp,pixels,width,maxval,format);
      pixels += width;
    }
    fclose(fp);
    if (stdcmap.base_pixel) {
      remappixels(data,width*height,stdcmap.base_pixel);
    }
    pxm = XCreatePixmap(dpy, RootWindow(dpy, screen), width, height, 8);
    if (!image) {
      image = XCreateImage(dpy,visual, 8,ZPixmap,0,0,width,height,8,0);
    }
    image->data = (char *) data;
    XPutImage(dpy, pxm, gc, image, 0,0,0,0,
          width,height);
    free(data);
    images[i++] = pxm;
  }

  attributes.override_redirect = True;
  attributes.bit_gravity = NorthWestGravity;
  attributes.event_mask =  VisibilityChangeMask;
  attributes.backing_store = NotUseful;
  attributes.colormap = colormap;

 image_win = XCreateWindow(dpy,
        RootWindow(dpy, screen),
        100, 100,
        width, height,
        0, 8, InputOutput, visual,
        CWBackingStore|CWColormap|CWEventMask|CWBitGravity,
        &attributes);


  sizehints.flags = PPosition | PSize | PMaxSize | USPosition;
  sizehints.width = width;
  sizehints.max_width = width;
  sizehints.height =  height;
  sizehints.max_height = height;
  sizehints.x = 10;
  sizehints.y = 10;
  XSetStandardProperties(dpy, image_win, "PGMMovie", "PGMMovie",
            None, argv, argc, &sizehints);


  XMapWindow(dpy, image_win);

  while (--repeats >= 0) {
    for (i = startframe ; i< argc; i++) {
      gettimeofday(&now,&dummy);
      last = timetoint(&now);
      XCopyArea(dpy, images[i], image_win, gc, 0, 0, width, height,  0, 0);

      while (XPending(dpy) > 0) {
        XNextEvent(dpy,&event);
      } 
      gettimeofday(&now,&dummy);
      left = 1000/framerate - 5 - (timetoint(&now) - last);
      if  (left > 0) {
        timeout.tv_sec = 0;
        timeout.tv_usec = 1000 * left;
        select(0,0,0,0,&timeout);
      }
    }
    if (linear) continue;
    for ( i = argc - 1; i > startframe ; --i) {
      gettimeofday(&now,&dummy);
      last = timetoint(&now);
      XCopyArea(dpy, images[i], image_win, gc, 0, 0, width, height,  0, 0);

      while (XPending(dpy) > 0) {
        XNextEvent(dpy,&event);
      } 
      gettimeofday(&now,&dummy);
      left = 1000/framerate - 5 - (timetoint(&now) - last);
      if  (left > 0) {
        timeout.tv_sec = 0;
        timeout.tv_usec = 1000 * left;
        select(0,0,0,0,&timeout);
      }
    }
  }
  return(0);
}
